@@ -45,6 +45,7 @@ struct gp_service {
     gss_cred_usage_t cred_usage;
     uint32_t filter_flags;
     uint32_t enforce_flags;
+    uint32_t min_lifetime;
     char *program;
 
     uint32_t mechs;
@@ -69,10 +70,23 @@ struct gp_config {
 struct gp_workers;
 
 struct gssproxy_ctx {
+    bool userproxymode;
+    char *config_file;
+    char *config_dir;
+    char *config_socket;
+    int daemonize;
+
     struct gp_config *config;
     struct gp_workers *workers;
     verto_ctx *vctx;
     verto_ev *sock_ev;      /* default socket event */
+
+    time_t term_timeout;
+    verto_ev *term_ev; /* termination ev in user mode */
+
+    ssize_t readstats;
+    ssize_t writestats;
+    time_t last_activity;
 };
 
 struct gp_sock_ctx {
@@ -97,13 +111,17 @@ struct gp_config *read_config(char *config_file, char *config_dir,
 struct gp_creds_handle *gp_service_get_creds_handle(struct gp_service *svc);
 void free_config(struct gp_config **config);
 void free_cred_store_elements(gss_key_value_set_desc *cs);
+struct gp_config *userproxy_config(char *socket_name, int opt_daemonize);
 
 /* from gp_init.c */
-void init_server(bool daemonize, int *wait_fd);
+void init_server(bool daemonize, int userproxy, int *wait_fd);
 void init_done(int wait_fd);
 void fini_server(void);
-verto_ctx *init_event_loop(void);
+int init_sockets(struct gssproxy_ctx *gpctx, struct gp_config *old_config);
+int init_userproxy_socket(struct gssproxy_ctx *gpctx);
+void init_event_loop(struct gssproxy_ctx *gpctx);
 void init_proc_nfsd(struct gp_config *cfg);
+int init_event_fini(struct gssproxy_ctx *gpctx);
 void write_pid(void);
 int drop_privs(struct gp_config *cfg);
 #ifdef HAVE_CAP
@@ -111,8 +129,20 @@ int drop_caps(void);
 int clear_bound_caps(void);
 #endif
 
+/* from gp_mgmt.c */
+void idle_handler(struct gssproxy_ctx *gpctx);
+void gp_activity_accounting(struct gssproxy_ctx *gpctx,
+                            ssize_t rb, ssize_t wb);
+void gp_krb5_tracing_setup(int action);
+void gp_krb5_fini_tracing(void);
+
 /* from gp_socket.c */
 void free_unix_socket(verto_ctx *ctx, verto_ev *ev);
+#ifdef HAVE_SYSTEMD_DAEMON
+int init_activation_socket(struct gssproxy_ctx *gpctx,
+                           const char *sock_name,
+                           struct gp_sock_ctx **sock_ctx);
+#endif
 struct gp_sock_ctx *init_unix_socket(struct gssproxy_ctx *gpctx,
                                      const char *file_name);
 void accept_sock_conn(verto_ctx *vctx, verto_ev *ev);
