@@ -10,6 +10,7 @@
 #include <fcntl.h>
 #include <netinet/in.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -17,6 +18,10 @@
 #include <sys/un.h>
 #include <unistd.h>
 
+#ifdef HAVE_SYSTEMD_DAEMON
+#include <systemd/sd-daemon.h>
+#endif
+
 #define FRAGMENT_BIT (1 << 31)
 
 struct unix_sock_conn {
@@ -167,9 +172,60 @@ void free_unix_socket(verto_ctx *ctx UNUSED, verto_ev *ev)
 {
     struct gp_sock_ctx *sock_ctx = NULL;
     sock_ctx = verto_get_private(ev);
+    close(sock_ctx->fd);
     free(sock_ctx);
 }
 
+#ifdef HAVE_SYSTEMD_DAEMON
+int init_activation_socket(struct gssproxy_ctx *gpctx,
+                           const char *sock_name,
+                           struct gp_sock_ctx **sock_ctx)
+{
+    int num_fds;
+    int ret = 0;
+
+    num_fds = sd_listen_fds(0);
+    if (num_fds > 1) {
+        GPDEBUG("Received too many socket from systemd\n");
+        ret = E2BIG;
+    } else if (num_fds == 1) {
+        struct gp_sock_ctx *_sock_ctx;
+        int fd = -1;
+
+        _sock_ctx = calloc(1, sizeof(struct gp_sock_ctx));
+        if (!_sock_ctx) {
+            ret = ENOMEM;
+            goto done;
+        }
+
+        fd = SD_LISTEN_FDS_START + 0;
+
+        ret = set_status_flags(fd, O_NONBLOCK);
+        if (ret != 0) {
+            GPDEBUG("Failed to set O_NONBLOCK on %d!\n", fd);
+            safefree(_sock_ctx);
+            goto done;
+        }
+
+        ret = set_fd_flags(fd, FD_CLOEXEC);
+        if (ret != 0) {
+            GPDEBUG("Failed to set FD_CLOEXEC on %d!\n", fd);
+            safefree(_sock_ctx);
+            goto done;
+        }
+
+        _sock_ctx->gpctx = gpctx;
+        _sock_ctx->socket = sock_name;
+        _sock_ctx->fd = fd;
+
+        *sock_ctx = _sock_ctx;
+    }
+
+done:
+    return ret;
+}
+#endif
+
 struct gp_sock_ctx *init_unix_socket(struct gssproxy_ctx *gpctx,
                                      const char *file_name)
 {
@@ -359,6 +415,7 @@ static void gp_socket_read(verto_ctx *vctx, verto_ev *ev)
     struct gp_buffer *rbuf;
     uint32_t size;
     bool header = false;
+    ssize_t total_read = 0;
     ssize_t rn;
     int ret;
     int fd;
@@ -386,6 +443,7 @@ static void gp_socket_read(verto_ctx *vctx, verto_ev *ev)
             ret = EIO;
             goto done;
         }
+        total_read = rn;
 
         /* allocate buffer for receiving data */
         rbuf->size = ntohl(size);
@@ -424,6 +482,7 @@ static void gp_socket_read(verto_ctx *vctx, verto_ev *ev)
         }
         goto done;
     }
+    total_read += rn;
 
     if (rn == 0) {
         if (!header) {
@@ -448,17 +507,22 @@ static void gp_socket_read(verto_ctx *vctx, verto_ev *ev)
 
         /* we successfully handed over the data */
         rbuf->data = NULL;
-        gp_buffer_free(rbuf);
-        return;
+        ret = 0;
+    } else {
+        ret = EAGAIN;
     }
 
-    ret = EAGAIN;
-
 done:
+    gp_activity_accounting(rbuf->conn->sock_ctx->gpctx,
+                           total_read, 0);
+
     switch (ret) {
     case EAGAIN:
         gp_socket_schedule_read(vctx, rbuf);
         return;
+    case 0:
+        gp_buffer_free(rbuf);
+        return;
     default:
         gp_conn_free(rbuf->conn);
         gp_buffer_free(rbuf);
@@ -508,6 +572,7 @@ static void gp_socket_write(verto_ctx *vctx, verto_ev *ev)
     struct gp_buffer *wbuf;
     struct iovec iov[2];
     uint32_t size;
+    ssize_t total_write = 0;
     ssize_t wn;
     int vecs;
     int fd;
@@ -549,6 +614,8 @@ static void gp_socket_write(verto_ctx *vctx, verto_ev *ev)
         }
         return;
     }
+    total_write = wn;
+
     if (vecs == 2) {
         if (wn < (ssize_t) sizeof(size)) {
             /* don't bother trying to handle sockets that can't
@@ -561,11 +628,15 @@ static void gp_socket_write(verto_ctx *vctx, verto_ev *ev)
         }
         wn -= sizeof(size);
     }
+    wbuf->pos += wn;
 
     GPDEBUGN(3, "[status] Sending data [%p (%zu)]: successful write of %d\n",
              wbuf->data, wbuf->size, wn);
 
-    wbuf->pos += wn;
+
+    gp_activity_accounting(wbuf->conn->sock_ctx->gpctx,
+                           0, total_write);
+
     if (wbuf->size > wbuf->pos) {
         /* short write, reschedule */
         gp_socket_schedule_write(vctx, wbuf);
